proc APSGenerateLinacElegantParameters {args} {
    set apsLinacMagnetsDir /home/helios/oagData/linac/magnets/
    set latticeDir /home/helios/oagData/linac/lattices/

    set gun RG2
    set gunMomentum 2.8
    set BCMomentum 150.0
    set postL4Momentum [expr 150+(425-150)/2.0]
    set postL5Momentum 425.0
    set input ""
    set output ""
    if {[APSStrictParseArguments {gun gunMomentum BCMomentum postL4Momentum postL5Momentum input output}] || \
          ![string length $input] || ![string length $output]} {
        return -code error "APSGenerateLinacElegantParameters: Invalid arguments $argv"
    }

    if ![file exists $input] {
        return -code error "APSGenerateLinacElegantParameters: not found: $input"
    }
    if [file exists $output] {
        return -code error "APSGenerateLinacElegantParameters: file in use: $output"
    }
    set inputColumnList [APSGetSDDSNames -fileName $input]
    foreach requiredColumn {ElementName Current} {
        if [lsearch -exact $inputColumnList $requiredColumn]==-1 {
            return -code error "APSGenerateLinacElegantParameters: $input lacks column $requiredColumn"
        }
    }
    
    set gunList [list RG1 RG2]
    if [lsearch -exact $gunList $gun]==-1 {
        return -code error "APSGenerateLinacElegantParameters: Invalid gun $gun. Valid choices: [join $gunList ,]"
    }
    set beamline ${gun}LINAC
    foreach position {gun BC postL4 postL5} {
        set momentum [set ${position}Momentum]
        if [expr $momentum<=0] {
            return -code error "APSGenerateLinacElegantParameters: ${position}Momentum is invalid ($momentum)"
        }
    }

    if ![file exists $latticeDir/linacMomentumProfileForParameterGeneration.ele] {
        return -code error "APSGenerateLinacElegantParameters: not found: $latticeDir/momentumProfile.ele"
    }

    set tmpRoot /tmp/[APSTmpString]
    if {[catch {exec elegant $latticeDir/linacMomentumProfileForParameterGeneration.ele \
                  -macro=latticeFile=$latticeDir/lattice.lte,beamline=$beamline \
                  -macro=latticeDir=$latticeDir \
                  -macro=p0=$gunMomentum,pPostL2=$BCMomentum,pPostL4=$postL4Momentum,pPostL5=$postL5Momentum \
                  -macro=rootname=$tmpRoot} result] || ![file exists $tmpRoot.done]} {
        return -code error "APSGenerateLinacElegantParameters: $result"
    }

    # Special handling for alpha magnets if present
    if [catch {exec sddsprocess $input -pipe=out -match=col,ElementName=*LFA* -nowarning \
                 | sddsxref $tmpRoot.cen -pipe -match=ElementName -take=pCentral -nowarning \
                 | sddsprocess -pipe \
                 "-print=column,ExcitationFile,$apsLinacMagnetsDir/%s.excitation,ElementName" \
                 | sddsinterpset -pipe \
                 -data=fileColumn=ExcitationFile,interpolate=B1,functionOf=Current,column=Current \
                 | sddsprocess -pipe=in -nowarning $tmpRoot.alpha \
                 -print=col,ElementParameter,X_MAX \
                 "-define=col,ParameterValue,pCentral B1 / sqrt 0.0750498 *"} result] {
        return -code error "APSGenerateLinacElegantParameters: $result"
    }

    if [catch {exec sddsprocess $tmpRoot.paramOpt $tmpRoot.rfParam -match=col,ElementType=RFCA} result] {
        return -code error "APSGenerateLinacElegantParameters: $result"
    }

    if [catch {exec sddsprocess $tmpRoot.paramOpt -pipe=out -match=col,ElementType=QUAD \
                 -match=col,ElementParameter=L \
                 | sddsbreak -pipe -change=ElementName \
                 | sddsprocess -pipe -process=ParameterValue,count,NSlices \
                 -process=ParameterValue,sum,LSum -process=ElementName,first,%s \
                 | sddscollapse -pipe=in $tmpRoot.lengths} result] {
        return -code error "APSGenerateLinacElegantParameters: $result"
    }

        
    if [catch {exec sddsprocess $tmpRoot.paramOpt -pipe=out \
                 -match=col,ElementType=*QUAD* -match=col,ElementParameter=L \
                 | sddsxref $tmpRoot.cen -pipe -match=ElementName -take=pCentral \
                 | sddsbreak -pipe -change=ElementName \
                 | sddsprocess -pipe -process=pCentral,ave,%sAve -process=ElementName,first,%s \
                 | sddscollapse -pipe \
                 | sddsxref -pipe $input -nowarning -match=ElementName -take=Current \
                 | sddsxref $tmpRoot.lengths -pipe -match=ElementName -take=LSum \
                 | sddsprocess -pipe \
                 "-print=column,ExcitationFile,$apsLinacMagnetsDir/%s.excitation,ElementName" \
                 | sddsinterpset -pipe \
                 -data=fileColumn=ExcitationFile,interpolate=IntegratedStrength,functionOf=Current,column=Current \
                 | sddsxref $apsLinacMagnetsDir/crossReference.sdds -pipe -match=ElementName -take=Sign \
                 | sddsprocess -pipe=in $tmpRoot.K1 \
                 -print=column,ElementParameter,K1 \
                 "-define=column,H,pCentralAve mev * 1e6 * c_mks /,units=T*m" \
                 "-define=col,ParameterValue,IntegratedStrength H / LSum / Sign *"} result] {
        return -code error "APSGenerateLinacElegantParameters: $result"
    }

    if [catch {exec sddscombine $tmpRoot.K1 $tmpRoot.alpha $tmpRoot.rfParam -merge -pipe=out \
                 -retain=col,pCentral*,ElementName,ElementParameter,ParameterValue,Current,ExcitationFile,IntegratedStrength \
                 | sddsprocess -pipe=in $output \
                 "-match=col,ElementParameter=FIDUCIAL,!,ElementParameter=BODY_FOCUS_MODEL,!,&"} result] {
        return -code error "APSGenerateLinacElegantParameters: $result"
    }

    eval file delete -force [glob $tmpRoot.*]
}

